@@ -35,7 +35,7 @@ static mini_cygheap NO_COPY cygheap_dummy =
 init_cygheap NO_COPY *cygheap = (init_cygheap *) &cygheap_dummy;
 void NO_COPY *cygheap_max;
 
-static NO_COPY muto cygheap_protect;
+static NO_COPY SRWLOCK cygheap_protect = SRWLOCK_INIT;
 
 struct cygheap_entry
 {
@@ -87,7 +87,8 @@ cygheap_fixup_in_child (bool execed)
   SIZE_T commit_size = CYGHEAP_STORAGE_INITIAL - CYGHEAP_STORAGE_LOW;
 
   if (child_proc_info->cygheap_max > (void *) CYGHEAP_STORAGE_INITIAL)
-    commit_size = allocsize (child_proc_info->cygheap_max);
+    commit_size = allocsize ((char *) child_proc_info->cygheap_max
+                   - CYGHEAP_STORAGE_LOW);
   cygheap = (init_cygheap *) VirtualAlloc ((LPVOID) CYGHEAP_STORAGE_LOW,
 					   CYGHEAP_STORAGE_HIGH
 					   - CYGHEAP_STORAGE_LOW,
@@ -129,6 +130,8 @@ void
 init_cygheap::close_ctty ()
 {
   debug_printf ("closing cygheap->ctty %p", cygheap->ctty);
+  if (cygheap->ctty->tc ()->getsid () == pid)
+    cygheap->ctty->tc ()->setsid (0); /* Release CTTY ownership */
   cygheap->ctty->close_with_arch ();
   cygheap->ctty = NULL;
 }
@@ -261,21 +264,20 @@ init_cygheap::init_installation_root ()
 
 /* Initialize bucket_val.  The value is the max size of a block
    fitting into the bucket.  The values are powers of two and their
-   medians: 24, 32, 48, 64, ...
+   medians: 32, 48, 64, 96, ...
    The idea is to have better matching bucket sizes (not wasting
    space) without trading in performance compared to the old powers
    of 2 method. */
 static const uint32_t bucket_val[NBUCKETS] = {
-  0, 24, 32, 48, 64, 96, 128, 192, 256, 384, 512, 768, 1024, 1536, 2048,
-  3072, 4096, 6144, 8192, 12288, 16384, 24576, 32768, 49152, 65536, 98304,
-  131072, 196608, 262144, 393216, 524288, 786432, 1048576, 1572864, 2097152,
-  3145728, 4194304, 6291456, 8388608, 12582912
+  32, 48, 64, 96, 128, 192, 256, 384,
+  512, 768, 1024, 1536, 2048, 3072, 4096, 6144,
+  8192, 12288, 16384, 24576, 32768, 49152, 65536, 98304,
+  131072, 196608, 262144, 393216, 524288, 786432, 1048576, 1572864
 };
 
 void
 cygheap_init ()
 {
-  cygheap_protect.init ("cygheap_protect");
   if (cygheap == &cygheap_dummy)
     {
       cygheap = (init_cygheap *) VirtualAlloc ((LPVOID) CYGHEAP_STORAGE_LOW,
@@ -291,7 +293,7 @@ cygheap_init ()
       cygheap->locale.mbtowc = __utf8_mbtowc;
       /* Set umask to a sane default. */
       cygheap->umask = 022;
-      cygheap->rlim_core = RLIM_INFINITY;
+      cygheap->rlim_core = 0;
     }
   if (!cygheap->fdtab)
     cygheap->fdtab.init ();
@@ -353,15 +355,19 @@ static void *
 _cmalloc (unsigned size)
 {
   _cmalloc_entry *rvc;
-  unsigned b;
+  unsigned b = 0;
 
   /* Calculate "bit bucket". */
-  for (b = 1; b < NBUCKETS && bucket_val[b] < size; b++)
-    continue;
+  if (size > bucket_val[0])
+    {
+      const unsigned clz = __builtin_clzl (size - 1);
+      b = (59 - clz) << 1;
+      b -= !((size - 1) & (1 << (62 - clz)));
+    }
   if (b >= NBUCKETS)
     return NULL;
 
-  cygheap_protect.acquire ();
+  AcquireSRWLockExclusive (&cygheap_protect);
   if (cygheap->buckets[b])
     {
       rvc = (_cmalloc_entry *) cygheap->buckets[b];
@@ -373,7 +379,7 @@ _cmalloc (unsigned size)
       rvc = (_cmalloc_entry *) _csbrk (bucket_val[b] + sizeof (_cmalloc_entry));
       if (!rvc)
 	{
-	  cygheap_protect.release ();
+	  ReleaseSRWLockExclusive (&cygheap_protect);
 	  return NULL;
 	}
 
@@ -381,19 +387,19 @@ _cmalloc (unsigned size)
       rvc->prev = cygheap->chain;
       cygheap->chain = rvc;
     }
-  cygheap_protect.release ();
+  ReleaseSRWLockExclusive (&cygheap_protect);
   return rvc->data;
 }
 
 static void
 _cfree (void *ptr)
 {
-  cygheap_protect.acquire ();
+  AcquireSRWLockExclusive (&cygheap_protect);
   _cmalloc_entry *rvc = to_cmalloc (ptr);
   unsigned b = rvc->b;
   rvc->ptr = cygheap->buckets[b];
   cygheap->buckets[b] = (char *) rvc;
-  cygheap_protect.release ();
+  ReleaseSRWLockExclusive (&cygheap_protect);
 }
 
 static void *
