@@ -1,8 +1,5 @@
 /* mount.cc: mount handling.
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
-   2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Red Hat, Inc.
-
 This file is part of Cygwin.
 
 This software is a copyrighted work licensed under the terms of the
@@ -38,6 +35,9 @@ details. */
    (path[mount_table->cygdrive_len + 1] == '/' || \
     !path[mount_table->cygdrive_len + 1]))
 
+#define isdev_disk(path) \
+  (path_prefix_p (dev_disk, (path), dev_disk_len, false))
+
 #define isproc(path) \
   (path_prefix_p (proc, (path), proc_len, false))
 
@@ -45,12 +45,34 @@ bool NO_COPY mount_info::got_usr_bin;
 bool NO_COPY mount_info::got_usr_lib;
 int NO_COPY mount_info::root_idx = -1;
 
-/* is_unc_share: Return non-zero if PATH begins with //server/share
+/* is_native_path: Return non-zero if PATH starts with \??\[a-zA-Z] or
+		   with \\?\[a-zA-Z] or with \\.\[a-zA-Z].
+
+   is_unc_share: Return non-zero if PATH begins with //server/share
 		 or with one of the native prefixes //./ or //?/
+
    This function is only used to test for valid input strings.
    The later normalization drops the native prefixes. */
 
-static inline bool __stdcall
+/* list of invalid chars in server names.  Note that underscore is ok,
+   but it cripples interoperability. */
+const char _invalid_server_char[] = ",~:!@#$%^&'.(){} ";
+#define valid_server_char(_c)	({				\
+		const char __c = (_c);				\
+		!iscntrl(__c)					\
+		&& strchr (_invalid_server_char, (_c)) == NULL;	\
+	})
+
+/* list of invalid chars in UNC filenames.  These are a few more than
+   for "normal" filenames. */
+const char _invalid_share_char[] = "\"/\\[]:|<>+=;,?*";
+#define valid_share_char(_c)	({				\
+		const char __c = (_c);				\
+		!iscntrl(__c)					\
+		&& strchr (_invalid_share_char, __c) == NULL;	\
+	})
+
+static inline bool
 is_native_path (const char *path)
 {
   return isdirsep (path[0])
@@ -60,15 +82,15 @@ is_native_path (const char *path)
 	 && isalpha (path[4]);
 }
 
-static inline bool __stdcall
+static inline bool
 is_unc_share (const char *path)
 {
   const char *p;
   return (isdirsep (path[0])
 	 && isdirsep (path[1])
-	 && isalnum (path[2])
+	 && valid_server_char (path[2])
 	 && ((p = strpbrk (path + 3, "\\/")) != NULL)
-	 && isalnum (p[1]));
+	 && valid_share_char (p[1]));
 }
 
 /* Return true if src_path is a valid, internally supported device name.
@@ -81,7 +103,7 @@ win32_device_name (const char *src_path, char *win32_path, device& dev)
   dev.parse (src_path);
   if (dev == FH_FS || dev == FH_DEV)
     return false;
-  strcpy (win32_path, dev.native);
+  strcpy (win32_path, dev.native ());
   return true;
 }
 
@@ -297,15 +319,52 @@ fs_info::update (PUNICODE_STRING upath, HANDLE in_vol)
   if (is_remote_drive ())
     {
 /* Should be reevaluated for each new OS.  Right now this mask is valid up
-   to Windows 8.  The important point here is to test only flags indicating
+   to Windows 11.  The important point here is to test only flags indicating
    capabilities and to ignore flags indicating a specific state of this
    volume.  At present these flags to ignore are FILE_VOLUME_IS_COMPRESSED,
-   FILE_READ_ONLY_VOLUME, and FILE_SEQUENTIAL_WRITE_ONCE.  The additional
-   filesystem flags supported since Windows 7 are also ignored for now.
-   They add information, but only on W7 and later, and only for filesystems
-   also supporting these flags, right now only NTFS. */
-#define GETVOLINFO_VALID_MASK (0x002701ffUL)
+   FILE_READ_ONLY_VOLUME, FILE_SEQUENTIAL_WRITE_ONCE and FILE_DAX_VOLUME. */
+#define GETVOLINFO_VALID_MASK (FILE_CASE_SENSITIVE_SEARCH		\
+			       | FILE_CASE_PRESERVED_NAMES		\
+			       | FILE_UNICODE_ON_DISK			\
+			       | FILE_PERSISTENT_ACLS			\
+			       | FILE_FILE_COMPRESSION			\
+			       | FILE_VOLUME_QUOTAS			\
+			       | FILE_SUPPORTS_SPARSE_FILES		\
+			       | FILE_SUPPORTS_REPARSE_POINTS		\
+			       | FILE_SUPPORTS_REMOTE_STORAGE		\
+			       | FILE_RETURNS_CLEANUP_RESULT_INFO	\
+			       | FILE_SUPPORTS_POSIX_UNLINK_RENAME	\
+			       | FILE_SUPPORTS_OBJECT_IDS		\
+			       | FILE_SUPPORTS_ENCRYPTION		\
+			       | FILE_NAMED_STREAMS			\
+			       | FILE_SUPPORTS_TRANSACTIONS		\
+			       | FILE_SUPPORTS_HARD_LINKS		\
+			       | FILE_SUPPORTS_EXTENDED_ATTRIBUTES	\
+			       | FILE_SUPPORTS_OPEN_BY_FILE_ID		\
+			       | FILE_SUPPORTS_USN_JOURNAL		\
+			       | FILE_SUPPORTS_INTEGRITY_STREAMS	\
+			       | FILE_SUPPORTS_BLOCK_REFCOUNTING	\
+			       | FILE_SUPPORTS_SPARSE_VDL		\
+			       | FILE_SUPPORTS_GHOSTING)
+/* This is the pre-Win7 mask used to recognize 3rd-party drivers.  We'll never
+   learn in time when those drivers start to support the new (har har) Win7 FS
+   flags.  */
+#define GETVOLINFO_NON_WIN_MASK (FILE_CASE_SENSITIVE_SEARCH		\
+			       | FILE_CASE_PRESERVED_NAMES		\
+			       | FILE_UNICODE_ON_DISK			\
+			       | FILE_PERSISTENT_ACLS			\
+			       | FILE_FILE_COMPRESSION			\
+			       | FILE_VOLUME_QUOTAS			\
+			       | FILE_SUPPORTS_SPARSE_FILES		\
+			       | FILE_SUPPORTS_REPARSE_POINTS		\
+			       | FILE_SUPPORTS_REMOTE_STORAGE		\
+			       | FILE_SUPPORTS_OBJECT_IDS		\
+			       | FILE_SUPPORTS_ENCRYPTION		\
+			       | FILE_NAMED_STREAMS			\
+			       | FILE_SUPPORTS_TRANSACTIONS)
+
 #define TEST_GVI(f,m) (((f) & GETVOLINFO_VALID_MASK) == (m))
+#define TEST_GVI_NON_WIN(f,m) (((f) & GETVOLINFO_NON_WIN_MASK) == (m))
 
 /* FIXME: This flag twist is getting awkward.  There should really be some
    other method.  Maybe we need mount flags to allow the user to fix file
@@ -316,21 +375,23 @@ fs_info::update (PUNICODE_STRING upath, HANDLE in_vol)
 #define SAMBA_IGNORE (FILE_VOLUME_QUOTAS \
 		      | FILE_SUPPORTS_OBJECT_IDS \
 		      | FILE_UNICODE_ON_DISK)
-#define FS_IS_SAMBA TEST_GVI(flags () & ~SAMBA_IGNORE, \
+#define FS_IS_SAMBA TEST_GVI_NON_WIN(flags () & ~SAMBA_IGNORE, \
 			     FILE_CASE_SENSITIVE_SEARCH \
 			     | FILE_CASE_PRESERVED_NAMES \
 			     | FILE_PERSISTENT_ACLS)
 /* Netapp DataOnTap. */
 #define NETAPP_IGNORE (FILE_SUPPORTS_SPARSE_FILES \
+		       | FILE_SUPPORTS_REPARSE_POINTS \
 		       | FILE_PERSISTENT_ACLS)
-#define FS_IS_NETAPP_DATAONTAP TEST_GVI(flags () & ~NETAPP_IGNORE, \
+#define FS_IS_NETAPP_DATAONTAP TEST_GVI_NON_WIN(flags () & ~NETAPP_IGNORE, \
 			     FILE_CASE_SENSITIVE_SEARCH \
 			     | FILE_CASE_PRESERVED_NAMES \
 			     | FILE_UNICODE_ON_DISK \
 			     | FILE_NAMED_STREAMS)
-/* These are the minimal flags supported by NTFS since Windows 2000.  Every
-   filesystem not supporting these flags is not a native NTFS.  We subsume
-   them under the filesystem type "cifs". */
+/* These are the minimal flags supported by NTFS since Windows 7, when
+   checking a remote NTFS filesystem.  Every filesystem not supporting these
+   flags is not a native NTFS.
+   We subsume them under the filesystem type "cifs". */
 #define MINIMAL_WIN_NTFS_FLAGS (FILE_CASE_SENSITIVE_SEARCH \
 				| FILE_CASE_PRESERVED_NAMES \
 				| FILE_UNICODE_ON_DISK \
@@ -341,13 +402,17 @@ fs_info::update (PUNICODE_STRING upath, HANDLE in_vol)
 				| FILE_SUPPORTS_REPARSE_POINTS \
 				| FILE_SUPPORTS_OBJECT_IDS \
 				| FILE_SUPPORTS_ENCRYPTION \
-				| FILE_NAMED_STREAMS)
+				| FILE_NAMED_STREAMS \
+				| FILE_SUPPORTS_HARD_LINKS \
+				| FILE_SUPPORTS_EXTENDED_ATTRIBUTES)
 #define FS_IS_WINDOWS_NTFS TEST_GVI(flags () & MINIMAL_WIN_NTFS_FLAGS, \
 				    MINIMAL_WIN_NTFS_FLAGS)
 /* These are the exact flags of a real Windows FAT/FAT32 filesystem.
+   Newer FAT32/exFAT support FILE_SUPPORTS_ENCRYPTION as well.
    Anything else is a filesystem faking to be FAT. */
 #define WIN_FAT_FLAGS (FILE_CASE_PRESERVED_NAMES | FILE_UNICODE_ON_DISK)
-#define FS_IS_WINDOWS_FAT  TEST_GVI(flags (), WIN_FAT_FLAGS)
+#define FAT_IGNORE (FILE_SUPPORTS_ENCRYPTION)
+#define FS_IS_WINDOWS_FAT  TEST_GVI(flags () & ~FAT_IGNORE, WIN_FAT_FLAGS)
 
       if ((flags () & FILE_SUPPORTS_OBJECT_IDS)
 	  && NT_SUCCESS (NtQueryVolumeInformationFile (vol, &io, &ffoi,
@@ -381,6 +446,8 @@ fs_info::update (PUNICODE_STRING upath, HANDLE in_vol)
 	is_cifs (!FS_IS_WINDOWS_FAT);
       /* Then check remote filesystems honest about their name. */
       if (!got_fs ()
+	  /* Microsoft exFAT */
+	  && !is_exfat (RtlEqualUnicodeString (&fsname, &ro_u_exfat, FALSE))
 	  /* Microsoft NFS needs distinct access methods for metadata. */
 	  && !is_nfs (RtlEqualUnicodeString (&fsname, &ro_u_nfs, FALSE))
 	  /* MVFS == Rational ClearCase remote filesystem.  Has a couple of
@@ -388,7 +455,7 @@ fs_info::update (PUNICODE_STRING upath, HANDLE in_vol)
 	     and stuff like that. */
 	  && !is_mvfs (RtlEqualUnicodePathPrefix (&fsname, &ro_u_mvfs, FALSE))
 	  /* NWFS == Novell Netware FS.  Broken info class, see below. */
-	  /* NcFsd == Novell Netware FS via own driver since Windows Vista. */
+	  /* NcFsd == Novell Netware FS via own driver. */
 	  && !is_nwfs (RtlEqualUnicodeString (&fsname, &ro_u_nwfs, FALSE))
 	  && !is_ncfsd (RtlEqualUnicodeString (&fsname, &ro_u_ncfsd, FALSE))
 	  /* UNIXFS == TotalNet Advanced Server (TAS).  Doesn't support
@@ -396,15 +463,11 @@ fs_info::update (PUNICODE_STRING upath, HANDLE in_vol)
 	  && !is_unixfs (RtlEqualUnicodeString (&fsname, &ro_u_unixfs, FALSE))
 	  /* AFSRDRFsd == Andrew File System.  Doesn't support DOS attributes.
 	     Only native symlinks are supported. */
-	  && !is_afs (RtlEqualUnicodeString (&fsname, &ro_u_afs, FALSE))
+	  && !is_afs (RtlEqualUnicodeString (&fsname, &ro_u_afs, FALSE)))
+	{
 	  /* PrlSF == Parallels Desktop File System.  Has a bug in
 	     FileNetworkOpenInformation, see below. */
-	  && !is_prlfs (RtlEqualUnicodeString (&fsname, &ro_u_prlfs, FALSE)))
-	{
-	  /* Known remote file system with buggy open calls.  Further
-	     explanation in fhandler.cc (fhandler_disk_file::open_fs). */
-	  is_sunwnfs (RtlEqualUnicodeString (&fsname, &ro_u_sunwnfs, FALSE));
-	  has_buggy_open (is_sunwnfs ());
+	  is_prlfs (RtlEqualUnicodeString (&fsname, &ro_u_prlfs, FALSE));
 	}
       if (got_fs ())
 	{
@@ -431,15 +494,12 @@ fs_info::update (PUNICODE_STRING upath, HANDLE in_vol)
 	     only support this if the filename is non-null and the handle is
 	     the handle to a directory. NcFsd IR10 is supposed to be ok. */
 	  has_buggy_reopen (is_netapp () || is_nwfs ());
-	  /* Netapp and Parallels Desktop FS have problems with the
-	     FileNetworkOpenInformation info class.  Netapp doesn't
-	     implement it at all, Parallels always returns a size of 0. */
-	  has_broken_fnoi (is_netapp () || is_prlfs ());
 	}
     }
   if (!got_fs ()
       && !is_ntfs (RtlEqualUnicodeString (&fsname, &ro_u_ntfs, FALSE))
       && !is_fat (RtlEqualUnicodePathPrefix (&fsname, &ro_u_fat, TRUE))
+      && !is_exfat (RtlEqualUnicodeString (&fsname, &ro_u_exfat, FALSE))
       && !is_refs (RtlEqualUnicodeString (&fsname, &ro_u_refs, FALSE))
       && !is_csc_cache (RtlEqualUnicodeString (&fsname, &ro_u_csc, FALSE))
       && is_cdrom (ffdi.DeviceType == FILE_DEVICE_CD_ROM))
@@ -459,21 +519,26 @@ fs_info::update (PUNICODE_STRING upath, HANDLE in_vol)
      except on Samba which handles Windows clients case insensitive.
 
      NFS doesn't set the FILE_CASE_SENSITIVE_SEARCH flag but is case
-     sensitive.
+     sensitive. */
+  caseinsensitive ((!(flags () & FILE_CASE_SENSITIVE_SEARCH) || is_samba ())
+		   && !is_nfs ());
 
-     UDF on NT 5.x is broken (at least) in terms of case sensitivity.
-     The UDF driver reports the FILE_CASE_SENSITIVE_SEARCH capability
-     but:
-     - Opening the root directory for query seems to work at first,
-       but the filenames in the directory listing are mutilated.
-     - When trying to open a file or directory case sensitive, the file
-       appears to be non-existant. */
-  caseinsensitive (((!(flags () & FILE_CASE_SENSITIVE_SEARCH) || is_samba ())
-		    && !is_nfs ())
-		   || (is_udf () && wincap.has_broken_udf ()));
+  /* Check for being an SSD */
+  if (!is_cdrom ())
+    {
+      /* Theoretically FileFsVolumeFlagsInformation would be sufficient,
+	 but apparently it's not exposed into userspace. */
+      FILE_FS_SECTOR_SIZE_INFORMATION ffssi;
+
+      status = NtQueryVolumeInformationFile (vol, &io, &ffssi, sizeof ffssi,
+					     FileFsSectorSizeInformation);
+      if (NT_SUCCESS (status))
+	is_ssd (!!(ffssi.Flags & SSINFO_FLAGS_NO_SEEK_PENALTY));
+    }
 
   if (!in_vol)
     NtClose (vol);
+
   fsi_cache.add (hash, this);
   return true;
 }
@@ -487,13 +552,13 @@ mount_info::create_root_entry (const PWCHAR root)
   sys_wcstombs (native_root, PATH_MAX, root);
   assert (*native_root != '\0');
   if (add_item (native_root, "/",
-		MOUNT_SYSTEM | MOUNT_BINARY | MOUNT_IMMUTABLE | MOUNT_AUTOMATIC)
+		MOUNT_SYSTEM | MOUNT_IMMUTABLE | MOUNT_AUTOMATIC)
       < 0)
     api_fatal ("add_item (\"%s\", \"/\", ...) failed, errno %d", native_root, errno);
   /* Create a default cygdrive entry.  Note that this is a user entry.
      This allows to override it with mount, unless the sysadmin created
      a cygdrive entry in /etc/fstab. */
-  cygdrive_flags = MOUNT_BINARY | MOUNT_NOPOSIX | MOUNT_CYGDRIVE;
+  cygdrive_flags = MOUNT_NOPOSIX | MOUNT_CYGDRIVE;
   strcpy (cygdrive, CYGWIN_INFO_CYGDRIVE_DEFAULT_PREFIX "/");
   cygdrive_len = strlen (cygdrive);
 }
@@ -506,7 +571,7 @@ mount_info::init (bool user_init)
   PWCHAR pathend;
   WCHAR path[PATH_MAX];
 
-  pathend = wcpcpy (path, cygheap->installation_root);
+  pathend = wcpcpy (path, cygheap->installation_root.Buffer);
   if (!user_init)
     create_root_entry (path);
 
@@ -522,41 +587,23 @@ mount_info::init (bool user_init)
       if (!got_usr_bin)
       {
 	stpcpy (p, "\\bin");
-	add_item (native, "/usr/bin",
-		  MOUNT_SYSTEM | MOUNT_BINARY | MOUNT_AUTOMATIC);
+	add_item (native, "/usr/bin", MOUNT_SYSTEM | MOUNT_AUTOMATIC);
       }
       if (!got_usr_lib)
       {
 	stpcpy (p, "\\lib");
-	add_item (native, "/usr/lib",
-		  MOUNT_SYSTEM | MOUNT_BINARY | MOUNT_AUTOMATIC);
+	add_item (native, "/usr/lib", MOUNT_SYSTEM | MOUNT_AUTOMATIC);
       }
     }
 }
 
-static void
-set_flags (unsigned *flags, unsigned val)
-{
-  *flags = val;
-  if (!(*flags & PATH_BINARY))
-    {
-      *flags |= PATH_TEXT;
-      debug_printf ("flags: text (%y)", *flags & (PATH_TEXT | PATH_BINARY));
-    }
-  else
-    {
-      *flags |= PATH_BINARY;
-      debug_printf ("flags: binary (%y)", *flags & (PATH_TEXT | PATH_BINARY));
-    }
-}
-
 int
 mount_item::build_win32 (char *dst, const char *src, unsigned *outflags, unsigned chroot_pathlen)
 {
   int n, err = 0;
   const char *real_native_path;
   int real_posix_pathlen;
-  set_flags (outflags, (unsigned) flags);
+  *outflags = flags;
   if (!cygheap->root.exists () || posix_pathlen != 1 || posix_path[0] != '/')
     {
       n = native_pathlen;
@@ -601,8 +648,6 @@ mount_info::conv_to_win32_path (const char *src_path, char *dst, device& dev,
 {
   bool chroot_ok = !cygheap->root.exists ();
 
-  MALLOC_CHECK;
-
   dev = FH_FS;
 
   *flags = 0;
@@ -629,12 +674,11 @@ mount_info::conv_to_win32_path (const char *src_path, char *dst, device& dev,
   /* See if this is a cygwin "device" */
   if (win32_device_name (src_path, dst, dev))
     {
-      *flags = MOUNT_BINARY;	/* FIXME: Is this a sensible default for devices? */
+      *flags = 0;
       rc = 0;
       goto out_no_chroot_check;
     }
 
-  MALLOC_CHECK;
   /* If the path is on a network drive or a //./ resp. //?/ path prefix,
      bypass the mount table.  If it's // or //MACHINE, use the netdrive
      device. */
@@ -643,7 +687,7 @@ mount_info::conv_to_win32_path (const char *src_path, char *dst, device& dev,
       if (!strchr (src_path + 2, '/'))
 	{
 	  dev = *netdrive_dev;
-	  set_flags (flags, PATH_BINARY);
+	  *flags = 0;
 	}
       else
 	{
@@ -652,26 +696,33 @@ mount_info::conv_to_win32_path (const char *src_path, char *dst, device& dev,
 	     are rather (warning, poetic description ahead) windows into the
 	     native Win32 world.  This also gives the user an elegant way to
 	     change the settings for those paths in a central place. */
-	  set_flags (flags, (unsigned) cygdrive_flags);
+	  *flags = cygdrive_flags;
 	}
       backslashify (src_path, dst, 0);
       /* Go through chroot check */
       goto out;
     }
+  if (isdev_disk (src_path))
+    {
+      dev = *dev_disk_dev;
+      *flags = 0;
+      strcpy (dst, src_path);
+      goto out;
+    }
   if (isproc (src_path))
     {
       dev = *proc_dev;
       dev = fhandler_proc::get_proc_fhandler (src_path);
       if (dev == FH_NADA)
 	return ENOENT;
-      set_flags (flags, PATH_BINARY);
+      *flags = 0;
       if (isprocsys_dev (dev))
 	{
 	  if (src_path[procsys_len])
 	    backslashify (src_path + procsys_len, dst, 0);
 	  else	/* Avoid empty NT path. */
 	    stpcpy (dst, "\\");
-	  set_flags (flags, (unsigned) cygdrive_flags);
+	  *flags = cygdrive_flags;
 	}
       else
 	strcpy (dst, src_path);
@@ -692,7 +743,7 @@ mount_info::conv_to_win32_path (const char *src_path, char *dst, device& dev,
 	}
       else if (cygdrive_win32_path (src_path, dst, unit))
 	{
-	  set_flags (flags, (unsigned) cygdrive_flags);
+	  *flags = cygdrive_flags;
 	  goto out;
 	}
       else if (mount_table->cygdrive_len > 1)
@@ -744,7 +795,6 @@ mount_info::conv_to_win32_path (const char *src_path, char *dst, device& dev,
       backslashify (src_path, dst + offset, 0);
     }
  out:
-  MALLOC_CHECK;
   if (chroot_ok || cygheap->root.ischroot_native (dst))
     rc = 0;
   else
@@ -759,12 +809,12 @@ mount_info::conv_to_win32_path (const char *src_path, char *dst, device& dev,
   return rc;
 }
 
-int
-mount_info::get_mounts_here (const char *parent_dir, int parent_dir_len,
+size_t
+mount_info::get_mounts_here (const char *parent_dir, size_t parent_dir_len,
 			     PUNICODE_STRING mount_points,
 			     PUNICODE_STRING cygd)
 {
-  int n_mounts = 0;
+  size_t n_mounts = 0;
 
   for (int i = 0; i < nmounts; i++)
     {
@@ -775,20 +825,29 @@ mount_info::get_mounts_here (const char *parent_dir, int parent_dir_len,
       if (last_slash == mi->posix_path)
 	{
 	  if (parent_dir_len == 1 && mi->posix_pathlen > 1)
-	    RtlCreateUnicodeStringFromAsciiz (&mount_points[n_mounts++],
-					      last_slash + 1);
+	    sys_mbstouni_alloc (&mount_points[n_mounts++], HEAP_BUF,
+			        last_slash + 1);
 	}
-      else if (parent_dir_len == last_slash - mi->posix_path
+      else if (parent_dir_len == (size_t) (last_slash - mi->posix_path)
 	       && strncasematch (parent_dir, mi->posix_path, parent_dir_len))
-	RtlCreateUnicodeStringFromAsciiz (&mount_points[n_mounts++],
-					  last_slash + 1);
+	sys_mbstouni_alloc (&mount_points[n_mounts++], HEAP_BUF,
+			    last_slash + 1);
     }
-  RtlCreateUnicodeStringFromAsciiz (cygd, cygdrive + 1);
+  sys_mbstouni_alloc (cygd, HEAP_BUF, cygdrive + 1);
   if (cygd->Length)
     cygd->Length -= 2;	// Strip trailing slash
   return n_mounts;
 }
 
+void
+mount_info::free_mounts_here (PUNICODE_STRING mount_points, int n_mounts,
+			      PUNICODE_STRING cygd)
+{
+  for (int i = 0; i < n_mounts; ++i)
+    cfree (mount_points[i].Buffer);
+  cfree (cygd->Buffer);
+}
+
 /* cygdrive_posix_path: Build POSIX path used as the
    mount point for cygdrives created when there is no other way to
    obtain a POSIX path from a Win32 one.
@@ -903,12 +962,12 @@ mount_info::conv_to_posix_path (const char *src_path, char *posix_path,
   else
     {
       const char *lastchar = src_path + src_path_len - 1;
-      append_slash = isdirsep (*lastchar) && lastchar[-1] != ':';
+      append_slash = isdirsep (*lastchar)
+		     && ((ccp_flags & __CCP_APP_SLASH) || lastchar[-1] != ':');
     }
 
   debug_printf ("conv_to_posix_path (%s, 0x%x, %s)", src_path, ccp_flags,
 		append_slash ? "add-slash" : "no-add-slash");
-  MALLOC_CHECK;
 
   if (src_path_len >= NT_MAX_PATH)
     {
@@ -962,7 +1021,7 @@ mount_info::conv_to_posix_path (const char *src_path, char *posix_path,
       if ((mi.posix_pathlen + (pathbuflen - mi.native_pathlen) + addslash) >= NT_MAX_PATH)
 	return ENAMETOOLONG;
       strcpy (posix_path, mi.posix_path);
-      if (addslash)
+      if (addslash || (!nextchar && append_slash))
 	strcat (posix_path, "/");
       if (nextchar)
 	slashify (p,
@@ -1011,25 +1070,9 @@ mount_info::conv_to_posix_path (const char *src_path, char *posix_path,
 
 out:
   debug_printf ("%s = conv_to_posix_path (%s)", posix_path, src_path);
-  MALLOC_CHECK;
   return 0;
 }
 
-/* Return flags associated with a mount point given the win32 path. */
-
-unsigned
-mount_info::set_flags_from_win32_path (const char *p)
-{
-  for (int i = 0; i < nmounts; i++)
-    {
-      mount_item &mi = mount[native_sorted[i]];
-      if (path_prefix_p (mi.native_path, p, mi.native_pathlen,
-			 mi.flags & MOUNT_NOPOSIX))
-	return mi.flags;
-    }
-  return PATH_BINARY;
-}
-
 inline char *
 skip_ws (char *in)
 {
@@ -1049,7 +1092,7 @@ find_ws (char *in)
 inline char *
 conv_fstab_spaces (char *field)
 {
-  register char *sp = field;
+  char *sp = field;
   while ((sp = strstr (sp, "\\040")) != NULL)
     {
       *sp++ = ' ';
@@ -1067,7 +1110,7 @@ struct opt
 {
   {"acl", MOUNT_NOACL, 1},
   {"auto", 0, 0},
-  {"binary", MOUNT_BINARY, 0},
+  {"binary", MOUNT_TEXT, 1},
   {"bind", MOUNT_BIND, 0},
   {"cygexec", MOUNT_CYGWIN_EXEC, 0},
   {"dos", MOUNT_DOS, 0},
@@ -1081,7 +1124,7 @@ struct opt
   {"posix=0", MOUNT_NOPOSIX, 0},
   {"posix=1", MOUNT_NOPOSIX, 1},
   {"sparse", MOUNT_SPARSE, 0},
-  {"text", MOUNT_BINARY, 1},
+  {"text", MOUNT_TEXT, 0},
   {"user", MOUNT_SYSTEM, 1}
 };
 
@@ -1183,7 +1226,7 @@ mount_info::from_fstab_line (char *line, bool user)
     return true;
   cend = find_ws (c);
   *cend = '\0';
-  unsigned mount_flags = MOUNT_SYSTEM | MOUNT_BINARY;
+  unsigned mount_flags = MOUNT_SYSTEM;
   if (!strcmp (fs_type, "cygdrive"))
     mount_flags |= MOUNT_NOPOSIX;
   if (!strcmp (fs_type, "usertemp"))
@@ -1200,7 +1243,7 @@ mount_info::from_fstab_line (char *line, bool user)
       int error = conv_to_win32_path (bound_path, native_path, dev, &flags);
       if (error || strlen (native_path) >= MAX_PATH)
 	return true;
-      if ((mount_flags & ~MOUNT_SYSTEM) == (MOUNT_BIND | MOUNT_BINARY))
+      if ((mount_flags & ~MOUNT_SYSTEM) == MOUNT_BIND)
 	mount_flags = (MOUNT_BIND | flags)
 		      & ~(MOUNT_IMMUTABLE | MOUNT_AUTOMATIC);
     }
@@ -1320,7 +1363,7 @@ mount_info::get_cygdrive_info (char *user, char *system, char *user_flags,
 	path[cygdrive_len - 1] = '\0';
     }
   if (flags)
-    strcpy (flags, (cygdrive_flags & MOUNT_BINARY) ? "binmode" : "textmode");
+    strcpy (flags, (cygdrive_flags & MOUNT_TEXT) ? "textmode" : "binmode");
   return 0;
 }
 
@@ -1328,8 +1371,8 @@ static mount_item *mounts_for_sort;
 
 /* sort_by_posix_name: qsort callback to sort the mount entries.  Sort
    user mounts ahead of system mounts to the same POSIX path. */
-/* FIXME: should the user should be able to choose whether to
-   prefer user or system mounts??? */
+/* FIXME: should the user be able to choose whether to prefer user or
+   system mounts??? */
 static int
 sort_by_posix_name (const void *a, const void *b)
 {
@@ -1364,8 +1407,8 @@ sort_by_posix_name (const void *a, const void *b)
 
 /* sort_by_native_name: qsort callback to sort the mount entries.  Sort
    user mounts ahead of system mounts to the same POSIX path. */
-/* FIXME: should the user should be able to choose whether to
-   prefer user or system mounts??? */
+/* FIXME: should the user be able to choose whether to prefer user or
+   system mounts??? */
 static int
 sort_by_native_name (const void *a, const void *b)
 {
@@ -1577,6 +1620,7 @@ mount_info::del_item (const char *path, unsigned flags)
 fs_names_t fs_names[] = {
     { "none", false },
     { "vfat", true },
+    { "exfat", true },
     { "ntfs", true },
     { "refs", true },
     { "smbfs", false },
@@ -1585,7 +1629,6 @@ fs_names_t fs_names[] = {
     { "iso9660", true },
     { "udf", true },
     { "csc-cache", false },
-    { "sunwnfs", false },
     { "unixfs", false },
     { "mvfs", false },
     { "cifs", false },
@@ -1602,14 +1645,8 @@ fillout_mntent (const char *native_path, const char *posix_path, unsigned flags)
   struct mntent& ret=_my_tls.locals.mntbuf;
   bool append_bs = false;
 
-  /* Remove drivenum from list if we see a x: style path */
   if (strlen (native_path) == 2 && native_path[1] == ':')
-    {
-      int drivenum = cyg_tolower (native_path[0]) - 'a';
-      if (drivenum >= 0 && drivenum <= 31)
-	_my_tls.locals.available_drives &= ~(1 << drivenum);
       append_bs = true;
-    }
 
   /* Pass back pointers to mount_table strings reserved for use by
      getmntent rather than pointers to strings in the internal mount
@@ -1646,7 +1683,7 @@ fillout_mntent (const char *native_path, const char *posix_path, unsigned flags)
      binary or textmode, or exec.  We don't print
      `silent' here; it's a magic internal thing. */
 
-  if (!(flags & MOUNT_BINARY))
+  if (flags & MOUNT_TEXT)
     strcpy (_my_tls.locals.mnt_opts, (char *) "text");
   else
     strcpy (_my_tls.locals.mnt_opts, (char *) "binary");
@@ -1701,33 +1738,34 @@ mount_item::getmntent ()
   return fillout_mntent (native_path, posix_path, flags);
 }
 
-static struct mntent *
-cygdrive_getmntent ()
+struct mntent *
+mount_info::cygdrive_getmntent ()
 {
-  char native_path[4];
-  char posix_path[CYG_MAX_PATH];
-  DWORD mask = 1, drive = 'a';
-  struct mntent *ret = NULL;
+  tmp_pathbuf tp;
+  const wchar_t *wide_path;
+  char *win32_path, *posix_path;
 
-  while (_my_tls.locals.available_drives)
+  if (!_my_tls.locals.drivemappings)
+    _my_tls.locals.drivemappings = new dos_drive_mappings (NO_FLOPPIES);
+
+  wide_path = _my_tls.locals.drivemappings->next_dos_mount ();
+  if (wide_path)
     {
-      for (/* nothing */; drive <= 'z'; mask <<= 1, drive++)
-	if (_my_tls.locals.available_drives & mask)
-	  break;
-
-      __small_sprintf (native_path, "%c:\\", cyg_toupper (drive));
-      if (GetFileAttributes (native_path) == INVALID_FILE_ATTRIBUTES)
+      win32_path = tp.c_get ();
+      sys_wcstombs (win32_path, NT_MAX_PATH, wide_path);
+      posix_path = tp.c_get ();
+      cygdrive_posix_path (win32_path, posix_path, 0);
+      return fillout_mntent (win32_path, posix_path, cygdrive_flags);
+    }
+  else
+    {
+      if (_my_tls.locals.drivemappings)
 	{
-	  _my_tls.locals.available_drives &= ~mask;
-	  continue;
+	  delete _my_tls.locals.drivemappings;
+	  _my_tls.locals.drivemappings = NULL;
 	}
-      native_path[2] = '\0';
-      __small_sprintf (posix_path, "%s%c", mount_table->cygdrive, drive);
-      ret = fillout_mntent (native_path, posix_path, mount_table->cygdrive_flags);
-      break;
+      return NULL;
     }
-
-  return ret;
 }
 
 struct mntent *
@@ -1802,7 +1840,7 @@ mount (const char *win32_path, const char *posix_path, unsigned flags)
 							   dev, &conv_flags);
 	      if (error || strlen (w32_path) >= MAX_PATH)
 		return true;
-	      if ((flags & ~MOUNT_SYSTEM) == (MOUNT_BIND | MOUNT_BINARY))
+	      if ((flags & ~MOUNT_SYSTEM) == MOUNT_BIND)
 		flags = (MOUNT_BIND | conv_flags)
 			& ~(MOUNT_IMMUTABLE | MOUNT_AUTOMATIC);
 	    }
@@ -1861,11 +1899,9 @@ cygwin_umount (const char *path, unsigned flags)
 #define is_dev(d,s)	wcsncmp((d),(s),sizeof(s) - 1)
 
 disk_type
-get_disk_type (LPCWSTR dos)
+get_device_type (LPCWSTR dev)
 {
-  WCHAR dev[MAX_PATH], *d = dev;
-  if (!QueryDosDeviceW (dos, dev, MAX_PATH))
-    return DT_NODISK;
+  const WCHAR *d = dev;
   if (is_dev (dev, L"\\Device\\"))
     {
       d += 8;
@@ -1896,18 +1932,24 @@ get_disk_type (LPCWSTR dos)
   return DT_NODISK;
 }
 
+disk_type
+get_disk_type (LPCWSTR dos)
+{
+  WCHAR dev[MAX_PATH];
+  if (!QueryDosDeviceW (dos, dev, MAX_PATH))
+    return DT_NODISK;
+  return get_device_type (dev);
+}
+
 extern "C" FILE *
 setmntent (const char *filep, const char *)
 {
   _my_tls.locals.iteration = 0;
-  _my_tls.locals.available_drives = GetLogicalDrives ();
-  /* Filter floppy drives on A: and B: */
-  if ((_my_tls.locals.available_drives & 1)
-      && get_disk_type (L"A:") == DT_FLOPPY)
-    _my_tls.locals.available_drives &= ~1;
-  if ((_my_tls.locals.available_drives & 2)
-      && get_disk_type (L"B:") == DT_FLOPPY)
-    _my_tls.locals.available_drives &= ~2;
+  if (_my_tls.locals.drivemappings)
+    {
+      delete _my_tls.locals.drivemappings;
+      _my_tls.locals.drivemappings = NULL;
+    }
   return (FILE *) filep;
 }
 
@@ -1951,81 +1993,181 @@ endmntent (FILE *)
   return 1;
 }
 
-dos_drive_mappings::dos_drive_mappings ()
+static bool
+resolve_dos_device (const wchar_t *dosname, wchar_t *devpath)
+{
+  if (QueryDosDeviceW (dosname, devpath, NT_MAX_PATH))
+    {
+      /* The DOS drive mapping can be another symbolic link.  If so,
+	 the mapping won't work since the section name is the name
+	 after resolving all symlinks.  Resolve symlinks here, too. */
+      for (int syml_cnt = 0; syml_cnt < SYMLOOP_MAX; ++syml_cnt)
+	{
+	  UNICODE_STRING upath;
+	  OBJECT_ATTRIBUTES attr;
+	  NTSTATUS status;
+	  HANDLE h;
+
+	  RtlInitUnicodeString (&upath, devpath);
+	  InitializeObjectAttributes (&attr, &upath, OBJ_CASE_INSENSITIVE,
+				      NULL, NULL);
+	  status = NtOpenSymbolicLinkObject (&h, SYMBOLIC_LINK_QUERY, &attr);
+	  if (!NT_SUCCESS (status))
+	    break;
+	  RtlInitEmptyUnicodeString (&upath, devpath, (NT_MAX_PATH - 1)
+						      * sizeof (WCHAR));
+	  status = NtQuerySymbolicLinkObject (h, &upath, NULL);
+	  NtClose (h);
+	  if (!NT_SUCCESS (status))
+	    break;
+	  devpath[upath.Length / sizeof (WCHAR)] = L'\0';
+	}
+      return true;
+    }
+  return false;
+}
+
+dos_drive_mappings::dos_drive_mappings (bool with_floppies)
 : mappings(0)
+, cur_mapping(0)
+, cur_dos(0)
 {
   tmp_pathbuf tp;
   wchar_t vol[64]; /* Long enough for Volume GUID string */
   wchar_t *devpath = tp.w_get ();
   wchar_t *mounts = tp.w_get ();
+  mapping **nextm = &mappings;
+  mapping *endfirstloop = NULL;
+  DWORD len;
 
-  /* Iterate over all volumes, fetch the first path from the list of
-     DOS paths the volume is mounted to, or use the GUID volume path
-     otherwise. */
-  HANDLE sh = FindFirstVolumeW (vol, 64);
-  if (sh == INVALID_HANDLE_VALUE)
-    debug_printf ("FindFirstVolumeW, %E");
-  else
-    do
+  /* Iterate over all drive letters, fetch the DOS device path */
+  if (!(len = GetLogicalDriveStringsW (NT_MAX_PATH - 1, mounts)) ||
+      len >= NT_MAX_PATH)
+    debug_printf ("GetLogicalDriveStringsW, %E");
+  else {
+    for (wchar_t *mount = mounts; *mount; mount += len + 2)
       {
-	/* Skip drives which are not mounted. */
-	DWORD len;
-	if (!GetVolumePathNamesForVolumeNameW (vol, mounts, NT_MAX_PATH, &len)
-	    || mounts[0] == L'\0')
-	  continue;
-	*wcsrchr (vol, L'\\') = L'\0';
-	if (QueryDosDeviceW (vol + 4, devpath, NT_MAX_PATH))
+	len = wcslen (mount);
+	mount[--len] = L'\0'; /* Drop trailing backslash */
+	if (resolve_dos_device (mount, devpath))
 	  {
-	    /* The DOS drive mapping can be another symbolic link.  If so,
-	       the mapping won't work since the section name is the name
-	       after resolving all symlinks.  Resolve symlinks here, too. */
-	    for (int syml_cnt = 0; syml_cnt < SYMLOOP_MAX; ++syml_cnt)
-	      {
-		UNICODE_STRING upath;
-		OBJECT_ATTRIBUTES attr;
-		NTSTATUS status;
-		HANDLE h;
-
-		RtlInitUnicodeString (&upath, devpath);
-		InitializeObjectAttributes (&attr, &upath,
-					    OBJ_CASE_INSENSITIVE, NULL, NULL);
-		status = NtOpenSymbolicLinkObject (&h, SYMBOLIC_LINK_QUERY,
-						   &attr);
-		if (!NT_SUCCESS (status))
-		  break;
-		RtlInitEmptyUnicodeString (&upath, devpath, (NT_MAX_PATH - 1)
-							    * sizeof (WCHAR));
-		status = NtQuerySymbolicLinkObject (h, &upath, NULL);
-		NtClose (h);
-		if (!NT_SUCCESS (status))
-		  break;
-		devpath[upath.Length / sizeof (WCHAR)] = L'\0';
-	      }
+	    if (!with_floppies && get_device_type (devpath) == DT_FLOPPY)
+	      continue;
 	    mapping *m = new mapping ();
 	    if (m)
 	      {
-		m->dospath = wcsdup (mounts);
+		m->dos.path = wcsdup (mount);
 		m->ntdevpath = wcsdup (devpath);
-		if (!m->dospath || !m->ntdevpath)
+		if (!m->dos.path || !m->ntdevpath)
 		  {
-		    free (m->dospath);
+		    free (m->dos.path);
 		    free (m->ntdevpath);
 		    delete m;
 		    continue;
 		  }
-		m->doslen = wcslen (m->dospath);
-		m->dospath[--m->doslen] = L'\0'; /* Drop trailing backslash */
+		m->dos.len = len;
 		m->ntlen = wcslen (m->ntdevpath);
-		m->next = mappings;
-		mappings = m;
+		*nextm = endfirstloop = m;
+		nextm = &m->next;
 	      }
 	  }
 	else
 	  debug_printf ("Unable to determine the native mapping for %ls "
-			"(error %u)", vol, GetLastError ());
+			"(error %E)", mount);
+      }
+  }
+
+  /* Iterate over all volumes, fetch the list of DOS paths the volume is
+     mounted to. */
+  HANDLE sh = FindFirstVolumeW (vol, 64);
+  if (sh == INVALID_HANDLE_VALUE)
+    debug_printf ("FindFirstVolumeW, %E");
+  else {
+    do
+      {
+	/* Skip volumes which are not mounted. */
+	if (!GetVolumePathNamesForVolumeNameW (vol, mounts, NT_MAX_PATH, &len)
+	    || mounts[0] == L'\0')
+	  continue;
+	/* Skip volumes which are only mounted to the root of a drive letter:
+	   they were handled in the loop above */
+	if (len == 5 && mounts[1] == L':' && mounts[2] == L'\\' && !mounts[3])
+	  continue;
+
+	*wcsrchr (vol, L'\\') = L'\0';
+	if (resolve_dos_device (vol + 4, devpath))
+	  {
+	    if (!with_floppies && get_device_type (devpath) == DT_FLOPPY)
+	      continue;
+	    mapping *m = new mapping ();
+	    bool hadrootmount = false;
+	    if (m)
+	      {
+		/* store mount point list */
+		if ((m->dos.path = (wchar_t *) malloc (len * sizeof (WCHAR))))
+		  memcpy (m->dos.path, mounts, len * sizeof (WCHAR));
+		m->ntdevpath = wcsdup (devpath);
+		if (!m->dos.path || !m->ntdevpath)
+		  {
+		    free (m->dos.path);
+		    free (m->ntdevpath);
+		    delete m;
+		    continue;
+		  }
+		/* split mount point list into dosmount entries */
+		mapping::dosmount *dos = &m->dos;
+		for (wchar_t *mount = m->dos.path;
+		    dos;
+		    mount += dos->len + 2,
+		      dos->next = mount[0] ? new mapping::dosmount () : NULL,
+		      dos = dos->next)
+		  {
+		    dos->path = mount;
+		    dos->len = wcslen (dos->path);
+		    dos->path[--dos->len] = L'\0'; /* Drop trailing backslash */
+		    if (dos->len == 2 && dos->path[1] == L':')
+		      hadrootmount = true;
+		  }
+		m->ntlen = wcslen (m->ntdevpath);
+		if (hadrootmount)
+		{
+		  /* This device has already been added to the mappings list
+		     in the first loop above, but with only the drive root
+		     mount.  Find that entry and replace it with the complete
+		     list of mounts. */
+		  hadrootmount = false;
+		  for (mapping *m2 = mappings;
+		       endfirstloop && m2 != endfirstloop->next;
+		       m2 = m2->next)
+		    {
+		      if (m->ntlen == m2->ntlen &&
+			  !wcscmp (m->ntdevpath, m2->ntdevpath))
+			{
+			  free (m2->dos.path);
+			  m2->dos.next = m->dos.next;
+			  m2->dos.path = m->dos.path;
+			  m2->dos.len = m->dos.len;
+			  free (m->ntdevpath);
+			  delete m;
+			  hadrootmount = true;
+			  break;
+			}
+		    }
+		}
+		if (!hadrootmount)
+		  {
+		    *nextm = m;
+		    nextm = &m->next;
+		  }
+	      }
+	  }
+	else
+	  debug_printf ("Unable to determine the native mapping for %ls "
+			"(error %E)", vol);
       }
     while (FindNextVolumeW (sh, vol, 64));
     FindVolumeClose (sh);
+  }
 }
 
 wchar_t *
@@ -2044,11 +2186,11 @@ dos_drive_mappings::fixup_if_match (wchar_t *path)
       {
 	wchar_t *tmppath;
 
-	if (m->ntlen > m->doslen)
-	  wcsncpy (path += m->ntlen - m->doslen, m->dospath, m->doslen);
+	if (m->ntlen > m->dos.len)
+	  wcsncpy (path += m->ntlen - m->dos.len, m->dos.path, m->dos.len);
 	else if ((tmppath = wcsdup (path + m->ntlen)) != NULL)
 	  {
-	    wcpcpy (wcpcpy (path, m->dospath), tmppath);
+	    wcpcpy (wcpcpy (path, m->dos.path), tmppath);
 	    free (tmppath);
 	  }
 	break;
@@ -2056,14 +2198,38 @@ dos_drive_mappings::fixup_if_match (wchar_t *path)
   return path;
 }
 
+const wchar_t *
+dos_drive_mappings::next_dos_mount ()
+{
+  if (cur_dos)
+    cur_dos = cur_dos->next;
+  while (!cur_dos)
+    {
+      if (cur_mapping)
+	cur_mapping = cur_mapping->next;
+      else
+	cur_mapping = mappings;
+      if (!cur_mapping)
+	return NULL;
+      cur_dos = &cur_mapping->dos;
+    }
+  return cur_dos->path;
+}
+
 dos_drive_mappings::~dos_drive_mappings ()
 {
   mapping *n = 0;
   for (mapping *m = mappings; m; m = n)
     {
       n = m->next;
-      free (m->dospath);
+      free (m->dos.path);
       free (m->ntdevpath);
+      mapping::dosmount *dn;
+      for (mapping::dosmount *dm = m->dos.next; dm; dm = dn)
+	{
+	  dn = dm->next;
+	  delete dm;
+	}
       delete m;
     }
 }
